#define _XOPEN_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>

#include "dsplit.c"

#define USAGE "Usage: %s <#inputs> <#iter> <seed>\n"

int main (int argc, char **argv)
{
  setbuf (stdout, NULL);

  if (argc != 4)
    {
      fprintf (stderr, USAGE, argv[0]);
      exit (1);
    }

#if DIR < 0
  if (fesetround (FE_DOWNWARD))
    fprintf (stderr, "%s: can't set rounding mode to RD\n", argv[0]);
#elif DIR > 0
  if (fesetround (FE_UPWARD))
    fprintf (stderr, "%s: can't set rounding mode to RU\n", argv[0]);
#endif

  long nmax =
    (LONG_MAX < (size_t) -1 ? LONG_MAX : (size_t) -1) / sizeof(num_t);
  long n = atol (argv[1]);
  if (n < 1 || n > nmax)
    {
      fprintf (stderr, "%s: the number of random inputs must be"
               " between 1 and %ld\n", argv[0], nmax);
      exit (1);
    }

  long k = atol (argv[2]);
  if (k < 1)
    {
      fprintf (stderr, "%s: the number of iterations must be at least 1\n",
               argv[0]);
      exit (1);
    }

  srand48 (strtol (argv[3], NULL, 0));

  num_t *t = (num_t *) malloc (n * sizeof(num_t));
  if (t == NULL)
    {
      fprintf (stderr, "%s: malloc failed\n", argv[0]);
      exit (1);
    }

  for (long i = 0; i < n; i++)
    {
      unsigned long long r = lrand48 ();
      r = (r << 31) + lrand48 ();
      unsigned long long msb = 1ULL << (PREC - 1);
      t[i] = (num_t) (msb | (r & (msb - 1)));
    }

  num_t d = 0;

  clock_t c = clock ();

  for (long j = 0; j < k; j++)
    for (long i = 0; i < n; i++)
      {
        dword_t res = split (t[i]);
        /* Make sure that all the results are taken into account, so that
           they are not discarded, and make sure that no optimizations can
           be done on the signs (the code will be inlined). Note: we need
           the product by res.h below to be potentially inexact, otherwise
           the negation in SplitRU could be delayed by the compiler as an
           unwanted optimization (this happens with Clang 9, and now the
           test shows that it is buggy with -O2 and -O3, e.g. on 1 1 1).
           By default, with -O3 and AVX, GCC vectorizes the code (length 4).
           In addition to the operations from split(), one has a vectorized
           products, vunpckhpd/vextractf128 instructions and 4 vaddsd, which
           introduce a overhead in the internal loop.
           Vectorization can be disabled with the -fno-tree-vectorize option.
           In this case, tiny differences in the timings can more likely be
           observed. */
        num_t z = res.l * res.l;
        d += res.h * z;
      }

  c = clock () - c;

  printf ("Sum = %a\n", (double) d);
  printf ("%.6f s\n", (double) c / CLOCKS_PER_SEC);

  free (t);

  return 0;
}
