#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include <mpfr.h>

static long maxfail = 0;
static int verbose = 0;

#define CHECK(E)                                \
  do {                                          \
    int _inex = (E);                            \
    assert (_inex == 0);                        \
  } while (0)

#define LOOP(X)                                 \
  for (mpfr_set_ui (X, 1, MPFR_RNDN);           \
       mpfr_get_exp (X) == 1;                   \
       mpfr_nextabove (X))

static char *vs[3] =
  {
    "RD(a + RD(a * c))",
    "RD(a + a * c)",
    "RD(a * RN(1 + c))",
  };

/* Split the input number a in rounding mode RNDD, and put the result
   in (ah,al).
   Return 0 if the required properties are satisfied, nonzero otherwise.
   Arguments:
   * ah and al have some precision p;
   * a and c should have the same precision p, but this does not matter;
   * t is a temporary variable that must be able to hold ah^2 and al^2,
     thus should have precision 2p;
   * p is the precision;
   * v describes the computation of a*:
     0: a* = RD(a + RD(a * c)), i.e. using a multiplication and an addition;
     1: a* = RD(a + a * c), i.e. using a FMA;
     2: a* = RD(a * c), where the constant c is denoted k in the paper
        (the constant provided in argument of the program, typically lambda,
        has been scaled by the expected power of 2, then 1 has been added to
        obtain c, which is denoted k in the paper).
*/
static int splitrd (mpfr_t ah, mpfr_t al, mpfr_t a, mpfr_t c, mpfr_t t,
                    int p, int v)
{
  mpfr_t f;
  mpfr_exp_t e, ea;
  int inex, s;

  ea = mpfr_get_exp (a);

  switch (v)
    {
    case 0:
      mpfr_mul (al, a, c, MPFR_RNDD);
      mpfr_add (ah, a, al, MPFR_RNDD);
      break;
    case 1:
      mpfr_fma (ah, a, c, a, MPFR_RNDD);
      break;
    case 2:
      mpfr_mul (ah, a, c, MPFR_RNDD);
      break;
    default:
      assert (0);
    }
  e = mpfr_get_exp (ah);  /* EXP(a*) */
  s = (p + 1) / 2;
  mpfr_mul_ui (al, ah, (1UL << s) + 1, MPFR_RNDD);  /* c */

  mpfr_init2 (f, p + s);
  CHECK (mpfr_sub (f, ah, al, MPFR_RNDD));  /* exact value of a* - c */

  mpfr_sub (ah, ah, al, MPFR_RNDD);  /* d */
  assert (mpfr_get_exp (ah) == e + s);

  CHECK (mpfr_sub (f, f, ah, MPFR_RNDD));  /* (a* - c) - d */
  assert (mpfr_sgn (f) >= 0);
  assert (mpfr_zero_p (f) || mpfr_get_exp (f) <= s + ea - p);
  /* i.e. 0 <= f < 2^s ulp(a) */
  mpfr_clear (f);

  mpfr_add (ah, ah, al, MPFR_RNDD);
  e = mpfr_get_exp (ah);
  /* Check that ah <= 2^EXP(a). */
  assert (e <= ea || (e == ea + 1 && mpfr_min_prec (ah) == 1));
  inex = mpfr_sub (al, a, ah, MPFR_RNDD);
  if (inex != 0)
    return 3;

  CHECK (mpfr_sqr (t, ah, MPFR_RNDN));
  if (mpfr_min_prec (t) > p)
    return 2;

  CHECK (mpfr_sqr (t, al, MPFR_RNDN));
  return mpfr_get_exp (t) > 2 * mpfr_get_exp (a) - p - ((p & 1) == 0);
}

static void test (mpfr_prec_t p, mpfr_t c, int v)
{
#define ALIST a, ah, al, almin, almin_a, almax, almax_a
  mpfr_t ALIST, t;
  long fail = 0;
  int w = -((p+13)/2 + (p >= 20));  /* for alignment in output */
  unsigned long almax_ul, albnd_ul;

  mpfr_inits2 (p, ALIST, (mpfr_ptr) 0);
  mpfr_init2 (t, 2 * p);

  mpfr_set_inf (almin, +1);
  mpfr_set_inf (almax, -1);

  LOOP (a)
    {
      int r = splitrd (ah, al, a, c, t, p, v);
      if (r)
        {
          if (verbose)
            {
              mpfr_printf ("Failure on %.*Rb (%d)\n", p - 1, a, r);
              mpfr_printf ("      ah = %.*Rb\n", p - 1, ah);
              mpfr_printf ("      al = %.*Rb\n", p - 1, al);
            }
          fail++;
          if (maxfail != 0 && fail >= maxfail)
            break;
        }
      if (mpfr_less_p (al, almin))
        {
          CHECK (mpfr_set (almin, al, MPFR_RNDN));
          CHECK (mpfr_set (almin_a, a, MPFR_RNDN));
        }
      if (mpfr_greater_p (al, almax))
        {
          CHECK (mpfr_set (almax, al, MPFR_RNDN));
          CHECK (mpfr_set (almax_a, a, MPFR_RNDN));
        }
    }

  /* Output should be aligned; a with full significand. */
  mpfr_printf ("almin = %+*Rb for a = %.*Rb\n", w, almin, p - 1, almin_a);
  mpfr_printf ("almax = %+*Rb for a = %.*Rb\n", w, almax, p - 1, almax_a);

  printf ("%ld failure(s)\n", fail);

  if (mpfr_cmpabs (almin, almax) > 0)
    CHECK (mpfr_abs (almax, almin, MPFR_RNDN));

  CHECK (mpfr_mul_2ui (almax, almax, p - 1, MPFR_RNDN));
  assert (mpfr_integer_p (almax));
  assert (mpfr_fits_ulong_p (almax, MPFR_RNDN));
  almax_ul = mpfr_get_ui (almax, MPFR_RNDN);
  assert (mpfr_cmp_ui (almax, almax_ul) == 0);

  /* Check whether |Al|max <= (2/3).2^(ceil(p/2)), which seems OK with
     c = 2/3 up to a power of 2, any precision, and any algorithm (v).
     Note: this is the bound from the paper without the 5/2 constant. */
  CHECK (mpfr_set_ui_2exp (t, 2, (p+1)/2, MPFR_RNDD));
  mpfr_div_ui (t, t, 3, MPFR_RNDD);
  assert (mpfr_fits_ulong_p (t, MPFR_RNDD));
  albnd_ul = mpfr_get_ui (t, MPFR_RNDD);
  printf ("|Al|max = %lu (bound %lu) %s\n",
          almax_ul, albnd_ul, almax_ul > albnd_ul ? "FAIL" : "OK");

  mpfr_clears (ALIST, t, (mpfr_ptr) 0);
}

static void readenv (void)
{
  char *s;

  s = getenv ("SPLITRD_MAXFAIL");
  if (s != NULL)
    maxfail = strtol (s, NULL, 0);

  s = getenv ("SPLITRD_VERBOSE");
  verbose = s != NULL && *s != '\0';

  printf ("maxfail = %ld, verbose = %d\n", maxfail, verbose);
}

int main (int argc, char **argv)
{
  long p, v;
  mpfr_t c;

  if (argc != 4)
    {
      fprintf (stderr, "Usage: %s <c> <prec> <variant>\n", argv[0]);
      exit (1);
    }

  p = strtol (argv[2], NULL, 0);
  if (p < 2 || p > sizeof (long) * CHAR_BIT)
    {
      fprintf (stderr, "%s: bad precision\n", argv[0]);
      exit (1);
    }

  v = strtol (argv[3], NULL, 0);
  if (v < 0 || v > 2)
    {
      fprintf (stderr, "%s: bad value of v, must be 0, 1 or 2\n", argv[0]);
      exit (1);
    }

  printf ("precision %ld, variant = %s\n", p, vs[v]);

  readenv ();

  mpfr_init2 (c, p);

  /* First argument: c, with corrected exponent. */
  if (mpfr_set_str (c, argv[1], 0, MPFR_RNDN))
    {
      fprintf (stderr, "%s: error when reading c\n", argv[0]);
      exit (1);
    }
  mpfr_set_exp (c, - (p / 2));
  mpfr_printf ("c = %.*Rb\n", p - 1, c);
  if (v == 2)
    {
      mpfr_add_ui (c, c, 1, MPFR_RNDN);  /* this is k in the paper */
      mpfr_printf ("k = %.*Rb\n", p - 1, c);
    }

  test (p, c, v);

  mpfr_clear (c);

  if (mpfr_nanflag_p ())
    {
      fprintf (stderr, "%s: a NaN occurred!\n", argv[0]);
      exit (1);
    }

  return 0;
}
